==> gpe-applauncher-0.9.tar.bz2: gpe-applauncher-0.9/hotkeys.c <==
/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * (c) kernel concepts 2007
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Application launcher system wide hotkey handler
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
/* gtk includes */
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gpe/desktop_file.h>

/* i8n */
#include <libintl.h>
#include <locale.h> 

#define _(x) gettext(x)

#include "gpe-applauncher.h"
#include "applications.h"
#include "dbus-if.h"

#define FILE_CONFIG  "/etc/gpe/hotkeys.conf"

typedef struct
{
  gchar *key;
  gchar *application;
  gboolean hold;
  gint keysym;
} t_hotkey;

typedef struct
{
  t_hotkey key_show_homescreen;
  t_hotkey key_toggle_homescreen;
  t_hotkey key_system_controls;
  gint appcount;
  t_hotkey *apps;
} t_keybindings;

static t_keybindings keybindings;


void
load_application_from_desktop (const gchar *name)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  gchar *filename;
	
  filename = g_strconcat (DEFAULT_APPS_PATH "/", name, ".desktop", NULL);
  
  p = gnome_desktop_file_load (filename, &err);
  g_free (filename);
  if (p)
    {
      gchar *name, *exec;
      gchar *type;
		
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type != NULL)
        {
          gnome_desktop_file_get_string (p, NULL, "Name", &name);
          gnome_desktop_file_get_string (p, NULL, "Exec", &exec);
		  launch_application (name, exec, "", TRUE, TRUE, SERVICE_APPLAUNCHER);
		  g_free (name);
		  g_free (exec);
        }
		
      gnome_desktop_file_free (p);
      if (type)
          g_free (type);
    }
  else
     fprintf (stderr, "Couldn't load \"%s\": %s\n", filename, err->message);

  if (err)
     g_error_free (err);
}


static GdkFilterReturn 
hotkeys_dispatch (GdkXEvent *xevent, GdkEvent *event, gpointer data)
{
  gint i;
  XEvent *ev = (XEvent*)xevent;

  if (ev->type == KeyPress)
    {
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_show_homescreen.keysym))
        {
          applauncher_present (TAB_HOMESCREEN);
          return GDK_FILTER_REMOVE;
        }
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_toggle_homescreen.keysym))
        {
          applauncher_toggle_home ();
          return GDK_FILTER_REMOVE;
        }
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_system_controls.keysym))
        {
		  /* to be implemented */
          return GDK_FILTER_REMOVE;
        }
      for (i = 0; i < keybindings.appcount; i++)
		{
		  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.apps[i].keysym))
			{
		      /* run app... */
              load_application_from_desktop (keybindings.apps[i].application);
              return GDK_FILTER_REMOVE;
			}
		}
    }

  if (event->type == GDK_BUTTON_RELEASE)
      XAllowEvents (GDK_DISPLAY(), SyncPointer, CurrentTime);

  return GDK_FILTER_CONTINUE;
}


static gboolean
grab_key (t_hotkey *hotkey)
{
  gint result; 
  int keycode;
	
  if (hotkey->key == NULL)
	 return FALSE;

  g_strstrip (hotkey->key);
  hotkey->keysym = gdk_keyval_from_name (hotkey->key);
  keycode = XKeysymToKeycode(GDK_DISPLAY(), hotkey->keysym);
  
  gdk_error_trap_push(); 
  
  XGrabKey (GDK_DISPLAY(), keycode, 0, GDK_ROOT_WINDOW(), False, 
            GrabModeAsync, GrabModeAsync);
  XSync (GDK_DISPLAY(), False);
    
  result = gdk_error_trap_pop();

  if (result)
    {
      if (result == BadAccess)
          g_printerr ("Unable to grab hotkey '%s', some program might own an active" \
                      "grab on it.", hotkey->key);
      else
          g_printerr ("Unable to grab the key '%s'.", hotkey->key);
	  return FALSE;
    }
  return TRUE;
}

static void
hotkeys_grab_keys (void)
{
  gint i;
	
  /* grab keys for builtin functions */
  grab_key (&keybindings.key_show_homescreen);
  grab_key (&keybindings.key_toggle_homescreen);
  grab_key (&keybindings.key_system_controls);

  /* grab application bindings */
  for (i = 1; i < keybindings.appcount; i++)
	grab_key (&keybindings.apps[i]);
}

static gboolean
hotkeys_load_configuration (void)
{
  GKeyFile *configfile;
  GError *err = NULL;
  gchar **keydefs, *akeydef;
  gint count = 0;

  configfile = g_key_file_new();
  		
  if (!g_key_file_load_from_file (configfile, FILE_CONFIG,
                                  G_KEY_FILE_KEEP_COMMENTS, &err))
	{
      fprintf(stderr, _("No hotkeys defined - disabling."));
	  g_error_free(err);
	  err = NULL;
      return FALSE;
	}
    
  keydefs = g_key_file_get_groups (configfile, NULL);
  if (keydefs == NULL)
      return FALSE;
  
  keybindings.apps = NULL;
  keybindings.appcount = 0;
	
  while (keydefs[count])
    {
      gboolean error_occured = FALSE;
        
      akeydef = keydefs[count];
		
      if (!strcmp (akeydef, "System"))
		{
          keybindings.key_show_homescreen.key = 
				g_key_file_get_string(configfile, akeydef, "homescreen_show", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
          keybindings.key_toggle_homescreen.key = 
				g_key_file_get_string(configfile, akeydef, "homescreen_toggle", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
          keybindings.key_system_controls.key = 
				g_key_file_get_string(configfile, akeydef, "system_controls", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
		}
	  else
		{
          keybindings.apps = g_realloc (keybindings.apps, sizeof (t_hotkey) * (keybindings.appcount + 1));
          memset (&keybindings.apps[keybindings.appcount], 0, sizeof (t_hotkey));
			
          keybindings.apps[keybindings.appcount].key = 
				g_key_file_get_string(configfile, akeydef, "key", &err);
			
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
          keybindings.apps[keybindings.appcount].application = 
				g_key_file_get_string(configfile, akeydef, "application", &err);
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
          keybindings.apps[keybindings.appcount].hold = 
				g_key_file_get_boolean(configfile, akeydef, "hold", &err);
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
		}
		
      count++;
	  if (error_occured)
		{
		  if (keybindings.apps[keybindings.appcount].key) 
			g_free (keybindings.apps[keybindings.appcount].key);
		  if (keybindings.apps[keybindings.appcount].application) 
			g_free (keybindings.apps[keybindings.appcount].application);
		}
	  else
		keybindings.appcount++;
	}
    
  g_strfreev(keydefs);
	
  return TRUE;
}

void
hotkeys_initialize (GtkWindow *window)
{
  hotkeys_load_configuration ();
  hotkeys_grab_keys ();
  gdk_window_add_filter (NULL, hotkeys_dispatch, NULL);
}

==> gpe-applauncher-0.10.tar.bz2: gpe-applauncher-0.10/hotkeys.c <==
/*
 * gpe-applauncher
 *
 * Application launcher for GPE Phone Edition.
 * (c) kernel concepts 2007
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Application launcher system wide hotkey handler
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
/* gtk includes */
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <gpe/desktop_file.h>

/* i8n */
#include <libintl.h>
#include <locale.h> 

#define _(x) gettext(x)

#include "gpe-applauncher.h"
#include "applications.h"
#include "dbus-if.h"
#include "controlpanel.h"
#include "launch-app.h"

#define FILE_CONFIG SYSCONFDIR "/gpe/hotkeys.conf"

typedef struct
{
  gchar *key;
  gchar *application;
  gboolean hold;
  gint keysym;
} t_hotkey;

typedef struct
{
  t_hotkey key_show_homescreen;
  t_hotkey key_toggle_homescreen;
  t_hotkey key_system_controls;
  gint appcount;
  t_hotkey *apps;
} t_keybindings;

static t_keybindings keybindings;


void
load_application_from_desktop (const gchar *name)
{
  GnomeDesktopFile *p;
  GError *err = NULL;
  gchar *filename;
	
  filename = g_strconcat (DEFAULT_APPS_PATH "/", name, ".desktop", NULL);
  
  p = gnome_desktop_file_load (filename, &err);
  g_free (filename);
  if (p)
    {
      gchar *name, *exec;
      gchar *type;
		
      gnome_desktop_file_get_string (p, NULL, "Type", &type);

      if (type != NULL)
        {
          gnome_desktop_file_get_string (p, NULL, "Name", &name);
          gnome_desktop_file_get_string (p, NULL, "Exec", &exec);
		  launch_application (name, exec, "", TRUE, TRUE, SERVICE_APPLAUNCHER);
		  g_free (name);
		  g_free (exec);
        }
		
      gnome_desktop_file_free (p);
      if (type)
          g_free (type);
    }
  else
     fprintf (stderr, "Couldn't load \"%s\": %s\n", filename, err->message);

  if (err)
     g_error_free (err);
}


static GdkFilterReturn 
hotkeys_dispatch (GdkXEvent *xevent, GdkEvent *event, gpointer data)
{
  gint i;
  XEvent *ev = (XEvent*)xevent;

  if (ev->type == KeyPress)
    {
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_show_homescreen.keysym))
        {
          applauncher_present (TAB_HOMESCREEN);
          return GDK_FILTER_REMOVE;
        }
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_toggle_homescreen.keysym))
        {
          applauncher_toggle_home ();
          return GDK_FILTER_REMOVE;
        }
	  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.key_system_controls.keysym))
        {
		  controlpanel_open (NULL);
          return GDK_FILTER_REMOVE;
        }
      for (i = 0; i < keybindings.appcount; i++)
		{
		  if (ev->xkey.keycode == XKeysymToKeycode(GDK_DISPLAY(), keybindings.apps[i].keysym))
			{
		      /* run app... */
              load_application_from_desktop (keybindings.apps[i].application);
              return GDK_FILTER_REMOVE;
			}
		}
    }

  if (event->type == GDK_BUTTON_RELEASE)
      XAllowEvents (GDK_DISPLAY(), SyncPointer, CurrentTime);

  return GDK_FILTER_CONTINUE;
}


static gboolean
grab_key (t_hotkey *hotkey)
{
  gint result; 
  int keycode;
	
  if (hotkey->key == NULL)
	 return FALSE;

  g_strstrip (hotkey->key);
  hotkey->keysym = gdk_keyval_from_name (hotkey->key);
  keycode = XKeysymToKeycode(GDK_DISPLAY(), hotkey->keysym);
  
  gdk_error_trap_push(); 
  
  XGrabKey (GDK_DISPLAY(), keycode, 0, GDK_ROOT_WINDOW(), False, 
            GrabModeAsync, GrabModeAsync);
  XSync (GDK_DISPLAY(), False);
    
  result = gdk_error_trap_pop();

  if (result)
    {
      if (result == BadAccess)
          g_printerr ("Unable to grab hotkey '%s', some program might own an active" \
                      "grab on it.", hotkey->key);
      else
          g_printerr ("Unable to grab the key '%s'.", hotkey->key);
	  return FALSE;
    }
  return TRUE;
}

static void
hotkeys_grab_keys (void)
{
  gint i;
	
  /* grab keys for builtin functions */
  grab_key (&keybindings.key_show_homescreen);
  grab_key (&keybindings.key_toggle_homescreen);
  grab_key (&keybindings.key_system_controls);

  /* grab application bindings */
  for (i = 1; i < keybindings.appcount; i++)
	grab_key (&keybindings.apps[i]);
}

static gboolean
hotkeys_load_configuration (void)
{
  GKeyFile *configfile;
  GError *err = NULL;
  gchar **keydefs, *akeydef;
  gint count = 0;

  configfile = g_key_file_new();
  		
  if (!g_key_file_load_from_file (configfile, FILE_CONFIG,
                                  G_KEY_FILE_KEEP_COMMENTS, &err))
	{
      fprintf(stderr, _("No hotkeys defined - disabling."));
	  g_error_free(err);
	  err = NULL;
      return FALSE;
	}
    
  keydefs = g_key_file_get_groups (configfile, NULL);
  if (keydefs == NULL)
      return FALSE;
  
  keybindings.apps = NULL;
  keybindings.appcount = 0;
	
  while (keydefs[count])
    {
      gboolean error_occured = FALSE;
        
      akeydef = keydefs[count];
		
      if (!strcmp (akeydef, "System"))
		{
          keybindings.key_show_homescreen.key = 
				g_key_file_get_string(configfile, akeydef, "homescreen_show", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
          keybindings.key_toggle_homescreen.key = 
				g_key_file_get_string(configfile, akeydef, "homescreen_toggle", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
          keybindings.key_system_controls.key = 
				g_key_file_get_string(configfile, akeydef, "system_controls", &err);
		  if (err != NULL)
			{
				g_error_free (err);
				err = NULL;
			}
		}
	  else
		{
          keybindings.apps = g_realloc (keybindings.apps, sizeof (t_hotkey) * (keybindings.appcount + 1));
          memset (&keybindings.apps[keybindings.appcount], 0, sizeof (t_hotkey));
			
          keybindings.apps[keybindings.appcount].key = 
				g_key_file_get_string(configfile, akeydef, "key", &err);
			
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
          keybindings.apps[keybindings.appcount].application = 
				g_key_file_get_string(configfile, akeydef, "application", &err);
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
          keybindings.apps[keybindings.appcount].hold = 
				g_key_file_get_boolean(configfile, akeydef, "hold", &err);
		  if (err != NULL)
			{
				error_occured = TRUE;
				g_error_free (err);
				err = NULL;
			}
		}
		
      count++;
	  if (error_occured)
		{
		  if (keybindings.apps[keybindings.appcount].key) 
			g_free (keybindings.apps[keybindings.appcount].key);
		  if (keybindings.apps[keybindings.appcount].application) 
			g_free (keybindings.apps[keybindings.appcount].application);
		}
	  else
		keybindings.appcount++;
	}
    
  g_strfreev(keydefs);
	
  return TRUE;
}

void
hotkeys_initialize (GtkWindow *window)
{
  hotkeys_load_configuration ();
  hotkeys_grab_keys ();
  gdk_window_add_filter (NULL, hotkeys_dispatch, NULL);
}
